'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

exports.default = function (_ref9) {
  var t = _ref9.types;
  var template = _ref9.template;

  function _ref(_id) {
    if (!Plugin(_id)) {
      throw new TypeError('Function return value violates contract.\n\nExpected:\nPlugin\n\nGot:\n' + _inspect(_id));
    }

    return _id;
  }

  if (!PluginParams(arguments[0])) {
    throw new TypeError('Value of argument 0 violates contract.\n\nExpected:\nPluginParams\n\nGot:\n' + _inspect(arguments[0]));
  }

  var PRESERVE_CONTEXTS = normalizeEnv(process.env.TRACE_CONTEXT);
  var PRESERVE_FILES = normalizeEnv(process.env.TRACE_FILE);
  var PRESERVE_LEVELS = normalizeEnv(process.env.TRACE_LEVEL);

  /**
   * Normalize an environment variable, used to override plugin options.
   */
  function normalizeEnv(input) {
    function _ref2(_id2) {
      if (!(Array.isArray(_id2) && _id2.every(function (item) {
        return typeof item === 'string';
      }))) {
        throw new TypeError('Function "normalizeEnv" return value violates contract.\n\nExpected:\nstring[]\n\nGot:\n' + _inspect(_id2));
      }

      return _id2;
    }

    if (!(input == null || typeof input === 'string')) {
      throw new TypeError('Value of argument "input" violates contract.\n\nExpected:\n?string\n\nGot:\n' + _inspect(input));
    }

    if (!input) {
      return _ref2([]);
    }
    return _ref2(input.split(/\s*,\s*/).map(function (context) {
      return context.toLowerCase().trim();
    }).filter(function (id) {
      return id;
    }));
  }

  /**
   * Like `template()` but returns an expression, not an expression statement.
   */
  function expression(input) {
    function _ref3(_id3) {
      if (!Template(_id3)) {
        throw new TypeError('Function "expression" return value violates contract.\n\nExpected:\nTemplate\n\nGot:\n' + _inspect(_id3));
      }

      return _id3;
    }

    if (!(typeof input === 'string')) {
      throw new TypeError('Value of argument "input" violates contract.\n\nExpected:\nstring\n\nGot:\n' + _inspect(input));
    }

    var fn = template(input);

    if (!Template(fn)) {
      throw new TypeError('Value of variable "fn" violates contract.\n\nExpected:\nTemplate\n\nGot:\n' + _inspect(fn));
    }

    return _ref3(function (ids) {
      function _ref4(_id4) {
        if (!Node(_id4)) {
          throw new TypeError('Function return value violates contract.\n\nExpected:\nNode\n\nGot:\n' + _inspect(_id4));
        }

        return _id4;
      }

      if (!TemplateIds(ids)) {
        throw new TypeError('Value of argument "ids" violates contract.\n\nExpected:\nTemplateIds\n\nGot:\n' + _inspect(ids));
      }

      var node = fn(ids);

      if (!Node(node)) {
        throw new TypeError('Value of variable "node" violates contract.\n\nExpected:\nNode\n\nGot:\n' + _inspect(node));
      }

      return _ref4(node.expression ? node.expression : node);
    });
  }

  /**
   * Normalize the plugin options.
   */
  function normalizeOpts(opts) {
    function _ref5(_id5) {
      if (!PluginOptions(_id5)) {
        throw new TypeError('Function "normalizeOpts" return value violates contract.\n\nExpected:\nPluginOptions\n\nGot:\n' + _inspect(_id5));
      }

      return _id5;
    }

    if (!PluginOptions(opts)) {
      throw new TypeError('Value of argument "opts" violates contract.\n\nExpected:\nPluginOptions\n\nGot:\n' + _inspect(opts));
    }

    if (opts[$normalized]) {
      return _ref5(opts);
    }

    // fill in the default handled labels if not specified
    if (!opts.labels) {
      opts.labels = ["error", "warn", "info", "debug", "trace"];

      if (!(Array.isArray(opts.labels) && opts.labels.every(function (item) {
        return typeof item === 'string';
      }))) {
        throw new TypeError('Value of "opts.labels" violates contract.\n\nExpected:\nArray<string>\n\nGot:\n' + _inspect(opts.labels));
      }
    }

    // add extra labels if specified
    if (opts.extraLabels) {
      opts.labels = opts.labels.concat(opts.extraLabels);

      if (!(Array.isArray(opts.labels) && opts.labels.every(function (item) {
        return typeof item === 'string';
      }))) {
        throw new TypeError('Value of "opts.labels" violates contract.\n\nExpected:\nArray<string>\n\nGot:\n' + _inspect(opts.labels));
      }
    }

    opts[$normalized] = true;
    return _ref5(opts);
  }

  function addIdentifiers(nodes) {
    return _extends({}, nodes, _lodash2.default.fromPairs(_lodash2.default.map(nodes, function (node, key) {
      return [key + 'Key', t.identifier(key)];
    })));
  }

  function logExpression(_ref10) {
    var logFunction = _ref10.logFunction;
    var content = _ref10.content;
    var metadata = _ref10.metadata;
    var label = _ref10.label;

    function _ref6(_id6) {
      if (!Node(_id6)) {
        throw new TypeError('Function "logExpression" return value violates contract.\n\nExpected:\nNode\n\nGot:\n' + _inspect(_id6));
      }

      return _id6;
    }

    return _ref6(expression('\n      ' + logFunction + '({\n        labelKey: label,\n        filenameKey: filename,\n        filepathKey: filepath,\n        contentKey: content,\n        lineKey: line,\n        parentPathKey: parentPath,\n      })\n    ')(addIdentifiers({
      label: t.stringLiteral(label),
      filename: t.stringLiteral(metadata.filename),
      filepath: t.stringLiteral(metadata.filepath),
      content: t.arrayExpression(t.isSequenceExpression(content) ? content.expressions : [content]),
      line: t.numericLiteral(metadata.line),
      parentPath: t.arrayExpression(_lodash2.default.map(metadata.parentPath, function (str) {
        if (!(typeof str === 'string')) {
          throw new TypeError('Value of argument "str" violates contract.\n\nExpected:\nstring\n\nGot:\n' + _inspect(str));
        }

        return t.stringLiteral(str);
      }))
    })));
  }

  /**
   * Collect the metadata for a given node path, which will be
   * made available to logging functions.
   */
  function collectMetadata(path, opts) {
    function _ref7(_id7) {
      if (!Metadata(_id7)) {
        throw new TypeError('Function "collectMetadata" return value violates contract.\n\nExpected:\nMetadata\n\nGot:\n' + _inspect(_id7));
      }

      return _id7;
    }

    if (!NodePath(path)) {
      throw new TypeError('Value of argument "path" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(path));
    }

    if (!PluginOptions(opts)) {
      throw new TypeError('Value of argument "opts" violates contract.\n\nExpected:\nPluginOptions\n\nGot:\n' + _inspect(opts));
    }

    var filename = path.hub.file.opts.filename;

    if (!(typeof filename === 'string')) {
      throw new TypeError('Value of variable "filename" violates contract.\n\nExpected:\nstring\n\nGot:\n' + _inspect(filename));
    }

    var filepath = _path2.default.resolve(process.cwd(), path.hub.file.opts.filename);

    if (!(typeof filepath === 'string')) {
      throw new TypeError('Value of variable "filepath" violates contract.\n\nExpected:\nstring\n\nGot:\n' + _inspect(filepath));
    }

    var line = path.node.loc.start.line;

    if (!(typeof line === 'number')) {
      throw new TypeError('Value of variable "line" violates contract.\n\nExpected:\nnumber\n\nGot:\n' + _inspect(line));
    }

    var indent = 0;

    var parent = void 0;

    if (!(parent == null || NodePath(parent))) {
      throw new TypeError('Value of variable "parent" violates contract.\n\nExpected:\n?NodePath\n\nGot:\n' + _inspect(parent));
    }

    var parentPath = path.getAncestry().slice(1).reduce(function (parts, item) {
      if (!(Array.isArray(parts) && parts.every(function (item) {
        return typeof item === 'string';
      }))) {
        throw new TypeError('Value of argument "parts" violates contract.\n\nExpected:\nstring[]\n\nGot:\n' + _inspect(parts));
      }

      if (!NodePath(item)) {
        throw new TypeError('Value of argument "item" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(item));
      }

      if (item.isClassMethod()) {
        if (!parent) {
          parent = item;
        }
        parts.unshift(item.node.key.type === 'Identifier' ? item.node.key.name : '[computed method]');
      } else if (item.isClassDeclaration()) {
        if (!parent) {
          parent = item;
        }
        parts.unshift(item.node.id ? item.node.id.name : '[anonymous class@' + item.node.loc.start.line + ']');
      } else if (item.isFunction()) {
        if (!parent) {
          parent = item;
        }
        parts.unshift(item.node.id && item.node.id.name || '[anonymous@' + item.node.loc.start.line + ']');
      } else if (item.isProgram()) {
        if (!parent) {
          parent = item;
        }
      } else if (!parent && !item.isClassBody() && !item.isBlockStatement()) {
        indent++;
      }
      return parts;
    }, []);

    if (!(Array.isArray(parentPath) && parentPath.every(function (item) {
      return typeof item === 'string';
    }))) {
      throw new TypeError('Value of variable "parentPath" violates contract.\n\nExpected:\nArray<string>\n\nGot:\n' + _inspect(parentPath));
    }

    var hasStartMessage = false;
    var isStartMessage = false;
    if (parent && !parent.isProgram()) {
      _parent$get$get = parent.get('body').get('body');

      if (!(_parent$get$get && (typeof _parent$get$get[Symbol.iterator] === 'function' || Array.isArray(_parent$get$get)))) {
        throw new TypeError('Expected _parent$get$get to be iterable, got ' + _inspect(_parent$get$get));
      }

      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = _parent$get$get[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          var _parent$get$get;

          var _child = _step.value;

          if (!NodePath(_child)) {
            throw new TypeError('Value of variable "child" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(_child));
          }

          if (_child.node[$handled]) {
            hasStartMessage = true;
            break;
          }
          if (!_child.isLabeledStatement()) {
            break;
          }
          var _label2 = _child.get('label');

          if (!NodePath(_label2)) {
            throw new TypeError('Value of variable "label" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(_label2));
          }

          if (_lodash2.default.includes(opts.labels, _label2.node.name)) {
            hasStartMessage = true;
            if (_child.node === path.node) {
              isStartMessage = true;
            }
            break;
          }
        }
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return) {
            _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }
    }

    return _ref7({
      indent: indent,
      parentPath: parentPath,
      hasStartMessage: hasStartMessage,
      isStartMessage: isStartMessage,
      filename: filename,
      filepath: filepath,
      line: line
    });
  }

  /**
   * Determine whether the given logging statement should be stripped.
   */
  function shouldStrip(name, metadata, opts) {
    if (!(typeof name === 'string')) {
      throw new TypeError('Value of argument "name" violates contract.\n\nExpected:\nstring\n\nGot:\n' + _inspect(name));
    }

    if (!Metadata(metadata)) {
      throw new TypeError('Value of argument "metadata" violates contract.\n\nExpected:\nMetadata\n\nGot:\n' + _inspect(metadata));
    }

    if (!PluginOptions(opts)) {
      throw new TypeError('Value of argument "opts" violates contract.\n\nExpected:\nPluginOptions\n\nGot:\n' + _inspect(opts));
    }

    if (!opts.strip) {
      return false;
    }

    if (
    // strip everything
    opts.strip === true ||

    // strip only a specific env (such as production)
    _lodash2.default.isString(opts.strip) && opts.strip === process.env.NODE_ENV ||

    // strip this specific env
    opts.strip.env && opts.strip.env[process.env.NODE_ENV] === true ||

    // strip this label
    opts.strip.labels && opts.strip.labels[name] === true) {
      return !hasStripOverride(name, metadata);
    }

    return false;
  }

  function hasStripOverride(name, metadata) {
    if (!(typeof name === 'string')) {
      throw new TypeError('Value of argument "name" violates contract.\n\nExpected:\nstring\n\nGot:\n' + _inspect(name));
    }

    if (!Metadata(metadata)) {
      throw new TypeError('Value of argument "metadata" violates contract.\n\nExpected:\nMetadata\n\nGot:\n' + _inspect(metadata));
    }

    if (PRESERVE_CONTEXTS.length && PRESERVE_CONTEXTS.some(function (context) {
      return metadata.context.toLowerCase().indexOf(context) !== -1;
    })) {
      return true;
    } else if (PRESERVE_FILES.length && PRESERVE_FILES.some(function (filename) {
      return metadata.filename.toLowerCase().indexOf(filename) !== -1;
    })) {
      return true;
    } else if (PRESERVE_LEVELS.length && PRESERVE_LEVELS.some(function (level) {
      return level === name.toLowerCase();
    })) {
      return true;
    } else {
      return false;
    }
  }

  return _ref({
    visitor: {
      Program: function Program(program, _ref11) {
        var opts = _ref11.opts;

        if (!NodePath(program)) {
          throw new TypeError('Value of argument "program" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(program));
        }

        program.traverse({
          LabeledStatement: function LabeledStatement(path) {
            if (!NodePath(path)) {
              throw new TypeError('Value of argument "path" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(path));
            }

            var label = path.get('label');

            if (!NodePath(label)) {
              throw new TypeError('Value of variable "label" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(label));
            }

            opts = normalizeOpts(opts);

            if (!_lodash2.default.includes(opts.labels, label.node.name)) {
              return;
            }

            var metadata = collectMetadata(path, opts);

            if (!Metadata(metadata)) {
              throw new TypeError('Value of variable "metadata" violates contract.\n\nExpected:\nMetadata\n\nGot:\n' + _inspect(metadata));
            }

            if (shouldStrip(label.node.name, metadata, opts)) {
              path.remove();
              return;
            }

            var logFunction = opts.logFunction || 'METALOG';

            path.traverse({
              "VariableDeclaration|Function|AssignmentExpression|UpdateExpression|YieldExpression|ReturnStatement": function VariableDeclarationFunctionAssignmentExpressionUpdateExpressionYieldExpressionReturnStatement(item) {
                if (!NodePath(item)) {
                  throw new TypeError('Value of argument "item" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(item));
                }

                throw path.buildCodeFrameError('Logging statements cannot have side effects.');
              },
              ExpressionStatement: function ExpressionStatement(statement) {
                if (!NodePath(statement)) {
                  throw new TypeError('Value of argument "statement" violates contract.\n\nExpected:\nNodePath\n\nGot:\n' + _inspect(statement));
                }

                if (statement.node[$handled]) {
                  return;
                }
                // const message: Message = {
                //   content: statement.get('expression').node,
                //   hasStartMessage: t.booleanLiteral(metadata.hasStartMessage),
                //   isStartMessage: t.booleanLiteral(metadata.isStartMessage),
                //   indent: t.numericLiteral(metadata.indent),
                //   parentName: t.stringLiteral(metadata.parentName),
                //   filename: t.stringLiteral(metadata.filename),
                //   filepath: t.stringLiteral(metadata.filepath),
                // };
                var content = statement.get('expression').node;

                if (!Node(content)) {
                  throw new TypeError('Value of variable "content" violates contract.\n\nExpected:\nNode\n\nGot:\n' + _inspect(content));
                }

                var replacement = t.expressionStatement(logExpression({
                  logFunction: logFunction,
                  content: content,
                  metadata: metadata,
                  label: label.node.name
                }));
                replacement[$handled] = true;
                statement.replaceWith(replacement);
              }
            });

            if (path.get('body').isBlockStatement()) {
              path.replaceWithMultiple(path.get('body').node.body);
            } else {
              path.replaceWith(path.get('body').node);
            }
          }
        });
      }
    }
  });
};

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var Plugin = function () {
  function Plugin(input) {
    return input != null && Visitors(input.visitor);
  }

  ;
  Object.defineProperty(Plugin, Symbol.hasInstance, {
    value: function value(input) {
      return Plugin(input);
    }
  });
  return Plugin;
}();

var PluginParams = function () {
  function PluginParams(input) {
    return input != null && input.types instanceof Object && typeof input.template === 'function';
  }

  ;
  Object.defineProperty(PluginParams, Symbol.hasInstance, {
    value: function value(input) {
      return PluginParams(input);
    }
  });
  return PluginParams;
}();

var PluginOptions = function () {
  function PluginOptions(input) {
    return input != null && (input.labels === undefined || Array.isArray(input.labels) && input.labels.every(function (item) {
      return typeof item === 'string';
    })) && (input.extraLabels === undefined || Array.isArray(input.extraLabels) && input.extraLabels.every(function (item) {
      return typeof item === 'string';
    })) && (input.strip === undefined || typeof input.strip === 'boolean' || typeof input.strip === 'string' || input.strip != null && (input.strip.env === undefined || input.strip.env != null && _typeof(input.strip.env) === 'object' && Object.keys(input.strip.env).every(function (key) {
      var _env = input.strip.env[key];
      return typeof _env === 'boolean';
    })) && (input.strip.labels === undefined || input.strip.labels != null && _typeof(input.strip.labels) === 'object' && Object.keys(input.strip.labels).every(function (key) {
      var _label = input.strip.labels[key];
      return typeof _label === 'boolean';
    })));
  }

  ;
  Object.defineProperty(PluginOptions, Symbol.hasInstance, {
    value: function value(input) {
      return PluginOptions(input);
    }
  });
  return PluginOptions;
}();

var Visitors = function () {
  function Visitors(input) {
    return input != null && (typeof input === 'undefined' ? 'undefined' : _typeof(input)) === 'object' && Object.keys(input).every(function (key) {
      var _key = input[key];
      return Visitor(_key);
    });
  }

  ;
  Object.defineProperty(Visitors, Symbol.hasInstance, {
    value: function value(input) {
      return Visitors(input);
    }
  });
  return Visitors;
}();

var Template = function () {
  function Template(input) {
    return typeof input === 'function';
  }

  ;
  Object.defineProperty(Template, Symbol.hasInstance, {
    value: function value(input) {
      return Template(input);
    }
  });
  return Template;
}();

var TemplateIds = function () {
  function TemplateIds(input) {
    return input != null && (typeof input === 'undefined' ? 'undefined' : _typeof(input)) === 'object' && Object.keys(input).every(function (key) {
      var _key2 = input[key];
      return Node(_key2);
    });
  }

  ;
  Object.defineProperty(TemplateIds, Symbol.hasInstance, {
    value: function value(input) {
      return TemplateIds(input);
    }
  });
  return TemplateIds;
}();

var Visitor = function () {
  function Visitor(input) {
    return typeof input === 'function';
  }

  ;
  Object.defineProperty(Visitor, Symbol.hasInstance, {
    value: function value(input) {
      return Visitor(input);
    }
  });
  return Visitor;
}();

var Node = function () {
  function Node(input) {
    return input != null && typeof input.type === 'string' && (input.node === undefined || input.node == null);
  }

  ;
  Object.defineProperty(Node, Symbol.hasInstance, {
    value: function value(input) {
      return Node(input);
    }
  });
  return Node;
}();

var Literal = function () {
  function Literal(input) {
    return input != null && (input.type === 'StringLiteral' || input.type === 'BooleanLiteral' || input.type === 'NumericLiteral' || input.type === 'NullLiteral' || input.type === 'RegExpLiteral');
  }

  ;
  Object.defineProperty(Literal, Symbol.hasInstance, {
    value: function value(input) {
      return Literal(input);
    }
  });
  return Literal;
}();

var Identifier = function () {
  function Identifier(input) {
    return input != null && typeof input.type === 'string' && typeof input.name === 'string';
  }

  ;
  Object.defineProperty(Identifier, Symbol.hasInstance, {
    value: function value(input) {
      return Identifier(input);
    }
  });
  return Identifier;
}();

var Scope = function () {
  function Scope(input) {
    return input != null && (typeof input === 'undefined' ? 'undefined' : _typeof(input)) === 'object';
  }

  ;
  Object.defineProperty(Scope, Symbol.hasInstance, {
    value: function value(input) {
      return Scope(input);
    }
  });
  return Scope;
}();

var NodePath = function () {
  function NodePath(input) {
    return input != null && typeof input.type === 'string' && Node(input.node) && Scope(input.scope);
  }

  ;
  Object.defineProperty(NodePath, Symbol.hasInstance, {
    value: function value(input) {
      return NodePath(input);
    }
  });
  return NodePath;
}();

var Metadata = function () {
  function Metadata(input) {
    return input != null && typeof input.indent === 'number' && Array.isArray(input.parentPath) && input.parentPath.every(function (item) {
      return typeof item === 'string';
    }) && typeof input.filename === 'string' && typeof input.filepath === 'string' && typeof input.hasStartMessage === 'boolean' && typeof input.isStartMessage === 'boolean' && typeof input.line === 'number';
  }

  ;
  Object.defineProperty(Metadata, Symbol.hasInstance, {
    value: function value(input) {
      return Metadata(input);
    }
  });
  return Metadata;
}();

var Level = function () {
  function Level(input) {
    return input === 'error' || input === 'warn' || input === 'info' || input === 'debug' || input === 'trace';
  }

  ;
  Object.defineProperty(Level, Symbol.hasInstance, {
    value: function value(input) {
      return Level(input);
    }
  });
  return Level;
}();

var $handled = Symbol('handled');
var $normalized = Symbol('normalized');

/**
 * # Trace
 */

function _inspect(input, depth) {
  var maxDepth = 4;
  var maxKeys = 15;

  if (depth === undefined) {
    depth = 0;
  }

  depth += 1;

  if (input === null) {
    return 'null';
  } else if (input === undefined) {
    return 'void';
  } else if (typeof input === 'string' || typeof input === 'number' || typeof input === 'boolean') {
    return typeof input === 'undefined' ? 'undefined' : _typeof(input);
  } else if (Array.isArray(input)) {
    if (input.length > 0) {
      var _ret = function () {
        if (depth > maxDepth) return {
            v: '[...]'
          };

        var first = _inspect(input[0], depth);

        if (input.every(function (item) {
          return _inspect(item, depth) === first;
        })) {
          return {
            v: first.trim() + '[]'
          };
        } else {
          return {
            v: '[' + input.slice(0, maxKeys).map(function (item) {
              return _inspect(item, depth);
            }).join(', ') + (input.length >= maxKeys ? ', ...' : '') + ']'
          };
        }
      }();

      if ((typeof _ret === 'undefined' ? 'undefined' : _typeof(_ret)) === "object") return _ret.v;
    } else {
      return 'Array';
    }
  } else {
    var keys = Object.keys(input);

    if (!keys.length) {
      if (input.constructor && input.constructor.name && input.constructor.name !== 'Object') {
        return input.constructor.name;
      } else {
        return 'Object';
      }
    }

    if (depth > maxDepth) return '{...}';

    var _indent = '  '.repeat(depth - 1);

    var entries = keys.slice(0, maxKeys).map(function (key) {
      return (/^([A-Z_$][A-Z0-9_$]*)$/i.test(key) ? key : JSON.stringify(key)) + ': ' + _inspect(input[key], depth) + ';';
    }).join('\n  ' + _indent);

    if (keys.length >= maxKeys) {
      entries += '\n  ' + _indent + '...';
    }

    if (input.constructor && input.constructor.name && input.constructor.name !== 'Object') {
      return input.constructor.name + ' {\n  ' + _indent + entries + '\n' + _indent + '}';
    } else {
      return '{\n  ' + _indent + entries + '\n' + _indent + '}';
    }
  }
}